"""SourceReader class for Cheetah's Parser and CodeGenerator
"""
import re
from .compat import unicode

EOLre = re.compile(r'[ \f\t]*(?:\r\n|\r|\n)')
EOLZre = re.compile(r'(?:\r\n|\r|\n|\Z)')
ENCODINGsearch = re.compile("coding[=:]\s*([-\w.]+)").search


class Error(Exception):
    pass


class SourceReader(object):
    def __init__(self, src, filename=None, breakPoint=None, encoding=None):
        self._src = src
        self._filename = filename
        self._srcLen = len(src)
        if breakPoint is None:
            self._breakPoint = self._srcLen
        else:
            self.setBreakPoint(breakPoint)
        self._pos = 0
        self._bookmarks = {}
        self._posTobookmarkMap = {}

        # collect some meta-information
        self._EOLs = []
        pos = 0
        while pos < len(self):
            EOLmatch = EOLZre.search(src, pos)
            self._EOLs.append(EOLmatch.start())
            pos = EOLmatch.end()

        self._BOLs = []
        for pos in self._EOLs:
            BOLpos = self.findBOL(pos)
            self._BOLs.append(BOLpos)

    def src(self):
        return self._src

    def filename(self):
        return self._filename

    def __len__(self):
        return self._breakPoint

    def __getitem__(self, i):
        if not isinstance(i, int):
            self.checkPos(i.stop)
        else:
            self.checkPos(i)
        return self._src[i]

    def __getslice__(self, i, j):
        i = max(i, 0)
        j = max(j, 0)
        return self._src[i:j]

    def splitlines(self):
        if not hasattr(self, '_srcLines'):
            self._srcLines = self._src.splitlines()
        return self._srcLines

    def lineNum(self, pos=None):
        if pos is None:
            pos = self._pos

        for i in range(len(self._BOLs)):
            if pos >= self._BOLs[i] and pos <= self._EOLs[i]:
                return i

    def getRowCol(self, pos=None):
        if pos is None:
            pos = self._pos
        lineNum = self.lineNum(pos)
        BOL = self._BOLs[lineNum]
        return lineNum + 1, pos - BOL + 1

    def getRowColLine(self, pos=None):
        if pos is None:
            pos = self._pos
        row, col = self.getRowCol(pos)
        return row, col, self.splitlines()[row - 1]

    def getLine(self, pos):
        if pos is None:
            pos = self._pos
        lineNum = self.lineNum(pos)
        return self.splitlines()[lineNum]

    def pos(self):
        return self._pos

    def setPos(self, pos):
        self.checkPos(pos)
        self._pos = pos

    def validPos(self, pos):
        return pos <= self._breakPoint and pos >= 0

    def checkPos(self, pos):
        if not pos <= self._breakPoint:
            raise Error(
                "pos (" + str(pos) + ") is invalid: beyond the stream's end ("
                + str(self._breakPoint - 1) + ")")
        elif not pos >= 0:
            raise Error("pos (" + str(pos) + ") is invalid: less than 0")

    def breakPoint(self):
        return self._breakPoint

    def setBreakPoint(self, pos):
        if pos > self._srcLen:
            raise Error(
                "New breakpoint (" + str(pos)
                + ") is invalid: beyond the end of stream's source string ("
                + str(self._srcLen) + ")")
        elif not pos >= 0:
            raise Error(
                "New breakpoint (" + str(pos) + ") is invalid: less than 0")

        self._breakPoint = pos

    def setBookmark(self, name):
        self._bookmarks[name] = self._pos
        self._posTobookmarkMap[self._pos] = name

    def hasBookmark(self, name):
        return name in self._bookmarks

    def gotoBookmark(self, name):
        if not self.hasBookmark(name):
            raise Error(
                "Invalid bookmark (" + name + ") is invalid: does not exist")
        pos = self._bookmarks[name]
        if not self.validPos(pos):
            raise Error("Invalid bookmark (" + name + ', '
                        + str(pos) + ") is invalid: pos is out of range")
        self._pos = pos

    def atEnd(self):
        return self._pos >= self._breakPoint

    def atStart(self):
        return self._pos == 0

    def peek(self, offset=0):
        self.checkPos(self._pos + offset)
        pos = self._pos + offset
        return self._src[pos]

    def getc(self):
        pos = self._pos
        if self.validPos(pos + 1):
            self._pos += 1
        return self._src[pos]

    def ungetc(self, c=None):
        if not self.atStart():
            raise Error('Already at beginning of stream')

        self._pos -= 1
        if c is not None:
            self._src[self._pos] = c

    def advance(self, offset=1):
        self.checkPos(self._pos + offset)
        self._pos += offset

    def rev(self, offset=1):
        self.checkPos(self._pos - offset)
        self._pos -= offset

    def read(self, offset):
        self.checkPos(self._pos + offset)
        start = self._pos
        self._pos += offset
        return self._src[start:self._pos]

    def readTo(self, to, start=None):
        self.checkPos(to)
        if start is None:
            start = self._pos
        self._pos = to
        return self._src[start:to]

    def readToEOL(self, start=None, gobble=True):
        EOLmatch = EOLZre.search(self.src(), self.pos())
        if gobble:
            pos = EOLmatch.end()
        else:
            pos = EOLmatch.start()
        return self.readTo(to=pos, start=start)

    def find(self, it, pos=None):
        if pos is None:
            pos = self._pos
        return self._src.find(it, pos)

    def startswith(self, it, pos=None):
        if self.find(it, pos) == self.pos():
            return True
        else:
            return False

    def rfind(self, it, pos):
        if pos is None:
            pos = self._pos
        return self._src.rfind(it, pos)

    def findBOL(self, pos=None):
        if pos is None:
            pos = self._pos
        src = self.src()
        return max(src.rfind('\n', 0, pos) + 1, src.rfind('\r', 0, pos) + 1, 0)

    def findEOL(self, pos=None, gobble=False):
        if pos is None:
            pos = self._pos

        match = EOLZre.search(self.src(), pos)
        if gobble:
            return match.end()
        else:
            return match.start()

    def isLineClearToPos(self, pos=None):
        if pos is None:
            pos = self.pos()
        self.checkPos(pos)
        src = self.src()
        BOL = self.findBOL()
        return BOL == pos or src[BOL:pos].isspace()

    def matches(self, strOrRE):
        if isinstance(strOrRE, (str, unicode)):
            return self.startswith(strOrRE, pos=self.pos())
        else:  # assume an re object
            return strOrRE.match(self.src(), self.pos())

    def matchWhiteSpace(self, WSchars=' \f\t'):
        return (not self.atEnd()) and self.peek() in WSchars

    def getWhiteSpace(self, max=None, WSchars=' \f\t'):
        if not self.matchWhiteSpace(WSchars):
            return ''
        start = self.pos()
        breakPoint = self.breakPoint()
        if max is not None:
            breakPoint = min(breakPoint, self.pos() + max)
        while self.pos() < breakPoint:
            self.advance()
            if not self.matchWhiteSpace(WSchars):
                break
        return self.src()[start:self.pos()]

    def matchNonWhiteSpace(self, WSchars=' \f\t\n\r'):
        return self.atEnd() or not self.peek() in WSchars

    def getNonWhiteSpace(self, WSchars=' \f\t\n\r'):
        if not self.matchNonWhiteSpace(WSchars):
            return ''
        start = self.pos()
        while self.pos() < self.breakPoint():
            self.advance()
            if not self.matchNonWhiteSpace(WSchars):
                break
        return self.src()[start:self.pos()]



#!/usr/bin/env python

import codecs
import glob
import os
import pprint
import re
import shutil
import sys
import json
try:
    import cPickle as pickle
except ImportError:  # PY3
    import pickle as pickle
from optparse import OptionParser

from .Compiler import DEFAULT_COMPILER_SETTINGS
from .Template import Template
from .Utils.Misc import mkdirsWithPyInitFiles
from .Version import Version
from .compat import PY2

optionDashesRE = re.compile(R"^-{1,2}")
moduleNameRE = re.compile(R"^[a-zA-Z_][a-zA-Z_0-9]*$")


def fprintfMessage(stream, format, *args):
    if format[-1:] == '^':
        format = format[:-1]
    else:
        format += '\n'
    if args:
        message = format % args
    else:
        message = format
    stream.write(message)


class Error(Exception):
    pass


class Bundle:
    """Wrap the source, destination and backup paths in one neat little class.
       Used by CheetahWrapper.getBundles().
    """
    def __init__(self, **kw):
        self.__dict__.update(kw)

    def __repr__(self):
        return "<Bundle %r>" % self.__dict__


##################################################
# USAGE FUNCTION & MESSAGES

def usage(usageMessage, errorMessage="", out=sys.stderr):
    """Write help text, an optional error message, and abort the program.
    """
    out.write(WRAPPER_TOP)
    out.write(usageMessage)
    exitStatus = 0
    if errorMessage:
        out.write('\n')
        out.write("*** USAGE ERROR ***: %s\n" % errorMessage)
        exitStatus = 1
    sys.exit(exitStatus)


WRAPPER_TOP = """\
         __  ____________  __
         \ \/            \/ /
          \/    *   *     \/    CHEETAH %(Version)s Command-Line Tool
           \      |       /
            \  ==----==  /      by Tavis Rudd <tavis@damnsimple.com>
             \__________/       and Mike Orr <sluggoster@gmail.com>

""" % globals()


HELP_PAGE1 = """\
USAGE:
------
  cheetah compile [options] [FILES ...]     : Compile template definitions
  cheetah fill [options] [FILES ...]        : Fill template definitions
  cheetah help                              : Print this help message
  cheetah options                           : Print options help message
  cheetah test [options]                    : Run Cheetah's regression tests
                                            : (same as for unittest)
  cheetah version                           : Print Cheetah version number

You may abbreviate the command to the first letter; e.g., 'h' == 'help'.
If FILES is a single "-", read standard input and write standard output.
Run "cheetah options" for the list of valid options.
"""

##################################################
# CheetahWrapper CLASS


class CheetahWrapper(object):
    MAKE_BACKUPS = True
    BACKUP_SUFFIX = ".bak"
    _templateClass = None
    _compilerSettings = None

    def __init__(self):
        self.progName = None
        self.command = None
        self.opts = None
        self.pathArgs = None
        self.sourceFiles = []
        self.searchList = []
        self.parser = None

    ##################################################
    # MAIN ROUTINE

    def main(self, argv=None):
        """The main program controller."""

        if argv is None:
            argv = sys.argv

        # Step 1: Determine the command and arguments.
        try:
            self.progName = os.path.basename(argv[0])
            self.command = command = optionDashesRE.sub("", argv[1])
            if command == 'test':
                self.testOpts = argv[2:]
            else:
                self.parseOpts(argv[2:])
        except IndexError:
            usage(HELP_PAGE1, "not enough command-line arguments")

        # Step 2: Call the command
        meths = (self.compile, self.fill, self.help, self.options,
                 self.test, self.version)
        for meth in meths:
            methName = meth.__name__
            # Or meth.__func__.__name__
            # Or meth.__name__
            methInitial = methName[0]
            if command in (methName, methInitial):
                sys.argv[0] += (" " + methName)
                # @@MO: I don't necessarily agree sys.argv[0] should be
                # modified.
                meth()
                return
        # If none of the commands matched.
        usage(HELP_PAGE1, "unknown command '%s'" % command)

    def parseOpts(self, args):
        D = self.debug
        self.isCompile = isCompile = self.command[0] == 'c'
        defaultOext = isCompile and ".py" or ".html"
        self.parser = OptionParser()
        pao = self.parser.add_option
        pao("--idir", action="store", dest="idir", default='',
            help='Input directory (defaults to current directory)')
        pao("--odir", action="store", dest="odir", default="",
            help='Output directory (defaults to current directory)')
        pao("--iext", action="store", dest="iext", default=".tmpl",
            help='File input extension '
                 '(defaults: compile: .tmpl, fill: .tmpl)')
        pao("--oext", action="store", dest="oext", default=defaultOext,
            help='File output extension (defaults: compile: .py, fill: .html)')
        pao("-R", action="store_true", dest="recurse", default=False,
            help='Recurse through subdirectories looking for input files')
        pao("--stdout", "-p", action="store_true", dest="stdout",
            default=False,
            help='Send output to stdout instead of writing to a file')
        pao("--quiet", action="store_false", dest="verbose", default=True,
            help='Do not print informational messages to stdout')
        pao("--debug", action="store_true", dest="debug", default=False,
            help='Print diagnostic/debug information to stderr')
        pao("--env", action="store_true", dest="env", default=False,
            help='Pass the environment into the search list')
        pao("--pickle", action="store", dest="pickle", default="",
            help='Unpickle FILE and pass it through in the search list')
        pao("--json", action="store", dest="json", default="",
            help='Read from JSON FILE and pass it through in the search list')
        pao("--flat", action="store_true", dest="flat", default=False,
            help='Do not build destination subdirectories')
        pao("--nobackup", action="store_true", dest="nobackup", default=False,
            help='Do not make backup files when generating new ones')
        pao("--settings", action="store", dest="compilerSettingsString",
            default=None,
            help='String of compiler settings to pass through, '
                 'e.g. --settings="useNameMapper=False,useFilters=False"')
        pao('--print-settings', action='store_true', dest='print_settings',
            help='Print out the list of available compiler settings')
        pao("--templateAPIClass", action="store", dest="templateClassName",
            default=None,
            help='Name of a subclass of Cheetah.Template.Template '
                 'to use for compilation, e.g. MyTemplateClass')
        pao("--parallel", action="store", type="int", dest="parallel",
            default=1,
            help='Compile/fill templates in parallel, e.g. --parallel=4')
        pao('--shbang', dest='shbang', default='#!/usr/bin/env python',
            help='Specify the shbang to place at the top '
                 'of compiled templates, e.g. --shbang="#!/usr/bin/python2.6"')
        pao('--encoding', dest='encoding', default=None,
            help='Specify the encoding of source files '
                 '(e.g. "utf-8" to force input files to be interpreted '
                 'as UTF-8)')

        opts, files = self.parser.parse_args(args)
        self.opts = opts
        if sys.platform == "win32":
            new_files = []
            for spec in files:
                file_list = glob.glob(spec)
                if file_list:
                    new_files.extend(file_list)
                else:
                    new_files.append(spec)
            files = new_files
        self.pathArgs = files

        D("""\
cheetah compile %s
Options are
%s
Files are %s""", args, pprint.pformat(vars(opts)), files)

        if opts.print_settings:
            print()
            print('>> Available Cheetah compiler settings:')
            from .Compiler import _DEFAULT_COMPILER_SETTINGS
            listing = _DEFAULT_COMPILER_SETTINGS
            listing.sort(key=lambda _l: _l[0][0].lower())

            for _l in listing:
                print('\t%s (default: "%s")\t%s' % _l)
            sys.exit(0)

        # cleanup trailing path separators
        seps = [sep for sep in [os.sep, os.altsep] if sep]
        for attr in ['idir', 'odir']:
            for sep in seps:
                path = getattr(opts, attr, None)
                if path and path.endswith(sep):
                    path = path[:-len(sep)]
                    setattr(opts, attr, path)
                    break

        self._fixExts()
        if opts.env:
            self.searchList.insert(0, os.environ)
        if opts.pickle:
            f = open(opts.pickle, 'rb')
            unpickled = pickle.load(f)
            f.close()
            self.searchList.insert(0, unpickled)
        if opts.json:
            f = open(opts.json, 'r')
            unjsoned = json.load(f)
            f.close()
            self.searchList.insert(0, unjsoned)

    ##################################################
    # COMMAND METHODS

    def compile(self):
        self._compileOrFill()

    def fill(self):
        from .ImportHooks import install
        install()
        self._compileOrFill()

    def help(self):
        usage(HELP_PAGE1, "", sys.stdout)

    def options(self):
        return self.parser.print_help()

    def test(self):
        # @@MO: Ugly kludge.
        TEST_WRITE_FILENAME = 'cheetah_test_file_creation_ability.tmp'
        try:
            f = open(TEST_WRITE_FILENAME, 'w')
        except Exception:
            sys.exit("""\
Cannot run the tests because you don't have write permission in the current
directory.  The tests need to create temporary files.  Change to a directory
you do have write permission to and re-run the tests.""")
        else:
            f.close()
            os.remove(TEST_WRITE_FILENAME)
        # @@MO: End ugly kludge.
        import unittest
        from .Tests import Test
        verbosity = 1
        if '-q' in self.testOpts:
            verbosity = 0
        if '-v' in self.testOpts:
            verbosity = 2
        runner = unittest.TextTestRunner(verbosity=verbosity)
        results = runner.run(unittest.TestSuite(Test.suites))
        exit(int(not results.wasSuccessful()))

    def version(self):
        print(Version)

    # If you add a command, also add it to the 'meths' variable in main().
    ##################################################
    # LOGGING METHODS

    def chatter(self, format, *args):
        """Print a verbose message to stdout.  But don't if .opts.stdout is
           true or .opts.verbose is false.
        """
        if self.opts.stdout or not self.opts.verbose:
            return
        fprintfMessage(sys.stdout, format, *args)

    def debug(self, format, *args):
        """Print a debugging message to stderr, but don't if .debug is
           false.
        """
        if self.opts.debug:
            fprintfMessage(sys.stderr, format, *args)

    def warn(self, format, *args):
        """Always print a warning message to stderr.
        """
        fprintfMessage(sys.stderr, format, *args)

    def error(self, format, *args):
        """
        Always print a warning message to stderr and exit with an error code.
        """
        fprintfMessage(sys.stderr, format, *args)
        sys.exit(1)

    ##################################################
    # HELPER METHODS

    def _fixExts(self):
        assert self.opts.oext, "oext is empty!"
        iext, oext = self.opts.iext, self.opts.oext
        if iext and not iext.startswith("."):
            self.opts.iext = "." + iext
        if oext and not oext.startswith("."):
            self.opts.oext = "." + oext

    def _compileOrFill(self):
        C, D = self.chatter, self.debug
        opts, files = self.opts, self.pathArgs
        if files == ["-"]:
            self._compileOrFillStdin()
            return
        elif not files and opts.recurse:
            which = opts.idir and "idir" or "current"
            C("Drilling down recursively from %s directory.", which)
            sourceFiles = []
            iext = opts.iext
            idir = os.path.join(opts.idir, os.curdir)
            for root, dirs, _files in os.walk(idir):
                for _f in _files:
                    _path = os.path.join(root, _f)
                    if _path.endswith(iext) and os.path.isfile(_path):
                        sourceFiles.append(_path)
        elif not files:
            usage(HELP_PAGE1, "Neither files nor -R specified!")
        else:
            sourceFiles = self._expandSourceFiles(files, opts.recurse, True)
        sourceFiles = [os.path.normpath(x) for x in sourceFiles]
        D("All source files found: %s", sourceFiles)
        bundles = self._getBundles(sourceFiles)
        D("All bundles: %s", pprint.pformat(bundles))
        if self.opts.flat:
            self._checkForCollisions(bundles)

        # In parallel mode a new process is forked for each template
        # compilation, out of a pool of size self.opts.parallel. This is not
        # really optimal in all cases (e.g. probably wasteful for small
        # templates), but seems to work well in real life for me.
        #
        # It also won't work for Windows users, but I'm not going to lose any
        # sleep over that.
        if self.opts.parallel > 1:
            bad_child_exit = 0
            pid_pool = set()

            def child_wait():
                pid, status = os.wait()
                pid_pool.remove(pid)
                return os.WEXITSTATUS(status)

            while bundles:
                b = bundles.pop()
                pid = os.fork()
                if pid:
                    pid_pool.add(pid)
                else:
                    self._compileOrFillBundle(b)
                    sys.exit(0)

                if len(pid_pool) == self.opts.parallel:
                    bad_child_exit = child_wait()
                    if bad_child_exit:
                        break

            while pid_pool:
                child_exit = child_wait()
                if not bad_child_exit:
                    bad_child_exit = child_exit

            if bad_child_exit:
                sys.exit("Child process failed, exited with code %d"
                         % bad_child_exit)

        else:
            for b in bundles:
                self._compileOrFillBundle(b)

    def _checkForCollisions(self, bundles):
        """Check for multiple source paths writing to the same destination
           path.
        """
        W = self.warn
        isError = False
        dstSources = {}
        for b in bundles:
            if b.dst in dstSources:
                dstSources[b.dst].append(b.src)
            else:
                dstSources[b.dst] = [b.src]
        keys = sorted(dstSources.keys())
        for dst in keys:
            sources = dstSources[dst]
            if len(sources) > 1:
                isError = True
                sources.sort()
                fmt = "Collision: multiple source files %s " \
                      "map to one destination file %s"
                W(fmt, sources, dst)
        if isError:
            what = self.isCompile and "Compilation" or "Filling"
            sys.exit("%s aborted due to collisions" % what)

    def _expandSourceFiles(self, files, recurse, addIextIfMissing):
        """Calculate source paths from 'files' by applying the
           command-line options.
        """
        D, W = self.debug, self.warn
        idir = self.opts.idir
        iext = self.opts.iext
        files = []
        for f in self.pathArgs:
            oldFilesLen = len(files)
            D("Expanding %s", f)
            path = os.path.join(idir, f)
            pathWithExt = path + iext  # May or may not be valid.
            if os.path.isdir(path):
                if recurse:
                    for root, dirs, _files in os.walk(path):
                        for _f in _files:
                            _path = os.path.join(root, _f)
                            if _path.endswith(iext) and os.path.isfile(_path):
                                files.append(_path)
                else:
                    raise Error("source file '%s' is a directory" % path)
            elif os.path.isfile(path):
                files.append(path)
            elif (addIextIfMissing and not path.endswith(iext)
                    and os.path.isfile(pathWithExt)):
                files.append(pathWithExt)
                # Do not recurse directories discovered by iext appending.
            elif os.path.exists(path):
                W("Skipping source file '%s', not a plain file.", path)
            else:
                W("Skipping source file '%s', not found.", path)
            if len(files) > oldFilesLen:
                D("  ... found %s", files[oldFilesLen:])
        return files

    def _getBundles(self, sourceFiles):
        flat = self.opts.flat
        idir = self.opts.idir
        iext = self.opts.iext
        odir = self.opts.odir
        oext = self.opts.oext
        idirSlash = idir + os.sep
        bundles = []
        for src in sourceFiles:
            # 'base' is the subdirectory plus basename.
            base = src
            if idir and src.startswith(idirSlash):
                base = src[len(idirSlash):]
            if iext and base.endswith(iext):
                base = base[:-len(iext)]
            basename = os.path.basename(base)
            if flat:
                dst = os.path.join(odir, basename + oext)
            else:
                dbn = basename
                if odir and base.startswith(os.sep):
                    odd = odir
                    while odd != '':
                        idx = base.find(odd)
                        if idx == 0:
                            dbn = base[len(odd):]
                            if dbn[0] == '/':
                                dbn = dbn[1:]
                            break
                        odd = os.path.dirname(odd)
                        if odd == '/':
                            break
                    dst = os.path.join(odir, dbn + oext)
                else:
                    dst = os.path.join(odir, base + oext)
            bak = dst + self.BACKUP_SUFFIX
            b = Bundle(src=src, dst=dst, bak=bak, base=base, basename=basename)
            bundles.append(b)
        return bundles

    def _getTemplateClass(self):
        C = self.chatter
        modname = None
        if self._templateClass:
            return self._templateClass

        modname = self.opts.templateClassName

        if not modname:
            return Template
        p = modname.rfind('.')
        if ':' not in modname:
            self.error('The value of option --templateAPIClass is invalid\n'
                       'It must be in the form "module:class", '
                       'e.g. "Cheetah.Template:Template"')

        modname, classname = modname.split(':')

        C('using --templateAPIClass=%s:%s' % (modname, classname))

        if p >= 0:
            mod = getattr(
                __import__(modname[:p], {}, {}, [modname[p+1:]]),  # noqa: E226,E501 missing whitespace around operator
                modname[p+1:])  # noqa: E226 missing whitespace around operator
        else:
            mod = __import__(modname, {}, {}, [])

        klass = getattr(mod, classname, None)
        if klass:
            self._templateClass = klass
            return klass
        else:
            self.error('**Template class specified '
                       'in option --templateAPIClass not found\n'
                       '**Falling back on Cheetah.Template:Template')

    def _getCompilerSettings(self):
        if self._compilerSettings:
            return self._compilerSettings

        def getkws(**kws):
            return kws
        if self.opts.compilerSettingsString:
            try:
                settings = eval('getkws(%s)'
                                % self.opts.compilerSettingsString)
            except Exception:
                self.error("There's an error in your --settings option."
                           "It must be valid Python syntax.\n"
                           + "    --settings='%s'\n"
                           % self.opts.compilerSettingsString
                           + "  %s: %s" % sys.exc_info()[:2]
                           )

            validKeys = set(DEFAULT_COMPILER_SETTINGS.keys())
            for k in settings:
                if k not in validKeys:
                    self.error(
                        'The --setting "%s" '
                        'is not a valid compiler setting name.'
                        % k)

            self._compilerSettings = settings
            return settings
        else:
            return {}

    def _compileOrFillStdin(self):
        TemplateClass = self._getTemplateClass()
        compilerSettings = self._getCompilerSettings()
        if self.isCompile:
            pysrc = TemplateClass.compile(file=sys.stdin,
                                          compilerSettings=compilerSettings,
                                          returnAClass=False)
            output = pysrc
        else:
            output = str(TemplateClass(file=sys.stdin,
                                       compilerSettings=compilerSettings))
        sys.stdout.write(output)

    def _compileOrFillBundle(self, b):
        C = self.chatter
        TemplateClass = self._getTemplateClass()
        compilerSettings = self._getCompilerSettings()
        src = b.src
        dst = b.dst
        basename = b.basename
        dstDir = os.path.dirname(dst)
        what = self.isCompile and "Compiling" or "Filling"
        C("%s %s -> %s^", what, src, dst)  # No trailing newline.
        if os.path.exists(dst) and not self.opts.nobackup:
            bak = b.bak
            C(" (backup %s)", bak)  # On same line as previous message.
        else:
            bak = None
            C("")
        if self.isCompile:
            if not moduleNameRE.match(basename):
                tup = basename, src
                raise Error("""\
%s: base name %s contains invalid characters.  It must
be named according to the same rules as Python modules.""" % tup)
            pysrc = TemplateClass.compile(file=src, returnAClass=False,
                                          moduleName=basename,
                                          className=basename,
                                          commandlineopts=self.opts,
                                          compilerSettings=compilerSettings)
            output = pysrc
        else:
            # output = str(TemplateClass(file=src, searchList=self.searchList))
            tclass = TemplateClass.compile(file=src,
                                           compilerSettings=compilerSettings)
            output = str(tclass(searchList=self.searchList))

        if bak:
            shutil.copyfile(dst, bak)
        if dstDir and not os.path.exists(dstDir):
            if self.isCompile:
                mkdirsWithPyInitFiles(dstDir)
            else:
                os.makedirs(dstDir)
        if self.opts.stdout:
            if not PY2:
                encoding = self.opts.encoding
                if encoding and isinstance(output, bytes):
                    output = output.decode(encoding)
            sys.stdout.write(output)
        else:
            encoding = self.opts.encoding
            if encoding:
                if isinstance(output, bytes):
                    output = output.decode(encoding)
                f = codecs.open(dst, 'w', encoding=encoding)
            else:
                if not PY2 and isinstance(output, bytes):
                    output = output.decode()
                f = open(dst, 'w')
            f.write(output)
            f.close()


# Called when invoked as `cheetah`
def _cheetah():
    CheetahWrapper().main()

# Called when invoked as `cheetah-compile`


def _cheetah_compile():
    sys.argv.insert(1, "compile")
    CheetahWrapper().main()


##################################################
# if run from the command line
if __name__ == '__main__':
    CheetahWrapper().main()



import os.path
import copy as copyModule
try:
    from ConfigParser import ConfigParser
except ImportError:  # PY3
    from configparser import ConfigParser
import re
from tokenize import Number
import types
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
from .compat import PY2


numberRE = re.compile(Number)
complexNumberRE = re.compile(
    '[\(]*' + Number + r'[ \t]*\+[ \t]*' + Number + '[\)]*')

##################################################
# FUNCTIONS ##


def mergeNestedDictionaries(dict1, dict2, copy=False, deepcopy=False):
    """Recursively merge the values of dict2 into dict1.

    This little function is very handy for selectively overriding settings in a
    settings dictionary that has a nested structure.
    """

    if copy:
        dict1 = copyModule.copy(dict1)
    elif deepcopy:
        dict1 = copyModule.deepcopy(dict1)

    for key, val in dict2.items():
        if key in dict1 and isinstance(val, dict) \
                and isinstance(dict1[key], dict):
            dict1[key] = mergeNestedDictionaries(dict1[key], val)
        else:
            dict1[key] = val
    return dict1


def stringIsNumber(S):
    """Return True if theString represents a Python number, False otherwise.
    This also works for complex numbers and numbers with +/- in front."""

    S = S.strip()

    if S[0] in '-+' and len(S) > 1:
        S = S[1:].strip()

    match = complexNumberRE.match(S)
    if not match:
        match = numberRE.match(S)
    if not match or (match.end() != len(S)):
        return False
    else:
        return True


def convStringToNum(theString):
    """
    Convert a string representation of a Python number to the Python version
    """
    if not stringIsNumber(theString):
        raise Error(theString + ' cannot be converted to a Python number')
    return eval(theString, {}, {})


class Error(Exception):
    pass


class NoDefault(object):
    pass


class ConfigParserCaseSensitive(ConfigParser):
    """A case sensitive version of the standard Python ConfigParser."""

    def optionxform(self, optionstr):
        """Don't change the case as is done in the default implemenation."""
        return optionstr


class _SettingsCollector(object):
    """An abstract base class that provides the methods SettingsManager uses to
    collect settings from config files and strings.

    This class only collects settings, it doesn't modify
    the _settings dictionary of SettingsManager instances in any way.
    """
    _ConfigParserClass = ConfigParserCaseSensitive

    def readSettingsFromModule(self, mod, ignoreUnderscored=True):
        """Returns all settings from a Python module.
        """
        S = {}
        attrs = vars(mod)
        for k, v in attrs.items():
            if (ignoreUnderscored and k.startswith('_')):
                continue
            else:
                S[k] = v
        return S

    def readSettingsFromPySrcStr(self, theString):
        """Return a dictionary of the settings in a Python src string."""

        globalsDict = {'True': (1 == 1),
                       'False': (0 == 1),
                       }
        newSettings = {'self': self}
        exec((theString + os.linesep), globalsDict, newSettings)
        del newSettings['self']
        module = types.ModuleType('temp_settings_module')
        module.__dict__.update(newSettings)
        return self.readSettingsFromModule(module)

    def readSettingsFromConfigFileObj(self, inFile, convert=True):
        """
        Return the settings from a config file that uses the syntax accepted by
        Python's standard ConfigParser module (like Windows .ini files).

        NOTE:
        this method maintains case unlike the ConfigParser module, unless this
        class was initialized with the 'caseSensitive' keyword set to False.

        All setting values are initially parsed as strings. However, If the
        'convert' arg is True this method will do the following value
        conversions:

        * all Python numeric literals will be coverted from string to number

        * The string 'None' will be converted to the Python value None

        * The string 'True' will be converted to a Python truth value

        * The string 'False' will be converted to a Python false value

        * Any string starting with 'python:' will be treated
          as a Python literal or expression that needs to be eval'd.
          This approach is useful for declaring lists and dictionaries.

        If a config section titled 'Globals' is present the options defined
        under it will be treated as top-level settings.
        """

        p = self._ConfigParserClass()
        if PY2:
            p.readfp(inFile)
        else:
            p.read_file(inFile)
        sects = p.sections()
        newSettings = {}

        sects = p.sections()
        newSettings = {}

        for s in sects:
            newSettings[s] = {}
            for o in p.options(s):
                if o != '__name__':
                    newSettings[s][o] = p.get(s, o)

        # loop through new settings -> deal with global settings, numbers,
        # booleans and None ++ also deal with 'importSettings' commands

        for sect, subDict in list(newSettings.items()):
            for key, val in list(subDict.items()):
                if convert:
                    if val.lower().startswith('python:'):
                        subDict[key] = eval(val[7:], {}, {})
                    if val.lower() == 'none':
                        subDict[key] = None
                    if val.lower() == 'true':
                        subDict[key] = True
                    if val.lower() == 'false':
                        subDict[key] = False
                    if stringIsNumber(val):
                        subDict[key] = convStringToNum(val)

                # now deal with any 'importSettings' commands
                if key.lower() == 'importsettings':
                    if val.find(';') < 0:
                        importedSettings = self.readSettingsFromPySrcFile(val)
                    else:
                        path = val.split(';')[0]
                        rest = ''.join(val.split(';')[1:]).strip()
                        parentDict = self.readSettingsFromPySrcFile(path)  # noqa: E501,F841
                        importedSettings = eval('parentDict["' + rest + '"]')

                    subDict.update(mergeNestedDictionaries(subDict,
                                                           importedSettings))

            if sect.lower() == 'globals':
                newSettings.update(newSettings[sect])
                del newSettings[sect]

        return newSettings


class SettingsManager(_SettingsCollector):
    """
    A mixin class that provides facilities for managing application settings.

    SettingsManager is designed to work well with nested settings dictionaries
    of any depth.
    """

    def __init__(self):
        super(SettingsManager, self).__init__()
        self._settings = {}
        self._initializeSettings()

    def _defaultSettings(self):
        return {}

    def _initializeSettings(self):
        """A hook that allows for complex setting initialization sequences
        that involve references to 'self' or other settings.  For example:
              self._settings['myCalcVal'] = self._settings['someVal'] * 15
        This method should be called by the class' __init__() method
        when needed.
        The dummy implementation should be reimplemented by subclasses.
        """
        pass

    # core post startup methods

    def setting(self, name, default=NoDefault):
        """
        Get a setting from self._settings, with or without a default value
        """

        if default is NoDefault:
            return self._settings[name]
        else:
            return self._settings.get(name, default)

    def hasSetting(self, key):
        """True/False"""
        return key in self._settings

    def setSetting(self, name, value):
        """Set a setting in self._settings."""
        self._settings[name] = value

    def settings(self):
        """Return a reference to the settings dictionary"""
        return self._settings

    def copySettings(self):
        """Returns a shallow copy of the settings dictionary"""
        return copyModule.copy(self._settings)

    def deepcopySettings(self):
        """Returns a deep copy of the settings dictionary"""
        return copyModule.deepcopy(self._settings)

    def updateSettings(self, newSettings, merge=True):
        """
        Update the settings with a selective merge or a complete overwrite
        """
        if merge:
            mergeNestedDictionaries(self._settings, newSettings)
        else:
            self._settings.update(newSettings)

    # source specific update methods

    def updateSettingsFromPySrcStr(self, theString, merge=True):
        """Update the settings from a code in a Python src string."""

        newSettings = self.readSettingsFromPySrcStr(theString)
        self.updateSettings(newSettings,
                            merge=newSettings.get('mergeSettings', merge))

    def updateSettingsFromConfigFileObj(self, inFile,
                                        convert=True, merge=True):
        """See the docstring for .updateSettingsFromConfigFile()

        The caller of this method is responsible for closing the inFile file
        object."""

        newSettings = self.readSettingsFromConfigFileObj(inFile,
                                                         convert=convert)
        self.updateSettings(newSettings,
                            merge=newSettings.get('mergeSettings', merge))

    def updateSettingsFromConfigStr(self, configStr, convert=True, merge=True):
        """See the docstring for .updateSettingsFromConfigFile()
        """

        configStr = '[globals]\n' + configStr
        inFile = StringIO(configStr)
        newSettings = self.readSettingsFromConfigFileObj(inFile,
                                                         convert=convert)
        self.updateSettings(newSettings,
                            merge=newSettings.get('mergeSettings', merge))



# -*- encoding: utf8 -*-

from glob import glob
import os
from shutil import rmtree
import tempfile
import unittest
from Cheetah.Compiler import Compiler
from Cheetah.Template import Template
from Cheetah import CheetahWrapper
from Cheetah.compat import PY2, unicode, load_module_from_file


class CommandLineTest(unittest.TestCase):
    def createAndCompile(self, source):
        fd, sourcefile = tempfile.mkstemp()
        os.close(fd)
        os.remove(sourcefile)
        sourcefile = sourcefile.replace('-', '_')

        if PY2:
            fd = open('%s.tmpl' % sourcefile, 'w')
        else:
            fd = open('%s.tmpl' % sourcefile, 'w', encoding='utf-8')
        fd.write(source)
        fd.close()

        wrap = CheetahWrapper.CheetahWrapper()
        wrap.main(['cheetah', 'compile',
                   '--encoding=utf-8', '--settings=encoding="utf-8"',
                   '--quiet', '--nobackup', sourcefile])
        module_name = os.path.split(sourcefile)[1]
        module = load_module_from_file(
            module_name, module_name, sourcefile + '.py')
        template = getattr(module, module_name)
        os.remove('%s.tmpl' % sourcefile)
        for sourcefile_py in glob('%s.py*' % sourcefile):  # *.py[co]
            os.remove(sourcefile_py)
        __pycache__ = os.path.join(os.path.dirname(sourcefile), '__pycache__')
        if os.path.exists(__pycache__):  # PY3
            rmtree(__pycache__)
        return template


class JBQ_UTF8_Test1(unittest.TestCase):
    def runTest(self):
        t = Template.compile(source="""Main file with |$v|

        $other""")

        otherT = Template.compile(source="Other template with |$v|")
        other = otherT()
        t.other = other

        t.v = u'Unicode String'
        t.other.v = u'Unicode String'

        assert unicode(t())


class JBQ_UTF8_Test2(unittest.TestCase):
    def runTest(self):
        t = Template.compile(source="""Main file with |$v|

        $other""")

        otherT = Template.compile(source="Other template with |$v|")
        other = otherT()
        t.other = other

        t.v = u'Unicode String with eacute é'
        t.other.v = u'Unicode String'

        assert unicode(t())


class JBQ_UTF8_Test3(unittest.TestCase):
    def runTest(self):
        t = Template.compile(source="""Main file with |$v|

        $other""")

        otherT = Template.compile(source="Other template with |$v|")
        other = otherT()
        t.other = other

        t.v = u'Unicode String with eacute é'
        t.other.v = u'Unicode String and an eacute é'

        assert unicode(t())


class JBQ_UTF8_Test4(unittest.TestCase):
    def runTest(self):
        t = Template.compile(source="""#encoding utf-8
        Main file with |$v| and eacute in the template é""")

        t.v = 'Unicode String'

        assert unicode(t())


class JBQ_UTF8_Test5(unittest.TestCase):
    def runTest(self):
        t = Template.compile(source="""#encoding utf-8
        Main file with |$v| and eacute in the template é""")

        t.v = u'Unicode String'

        assert unicode(t())


class JBQ_UTF8_Test6(unittest.TestCase):
    def runTest(self):
        source = """#encoding utf-8
        #set $someUnicodeString = u"Bébé"
        Main file with |$v| and eacute in the template é"""
        t = Template.compile(source=source)

        t.v = u'Unicode String'

        assert unicode(t())


class JBQ_UTF8_Test7(CommandLineTest):
    def runTest(self):
        source = """#encoding utf-8
        #set $someUnicodeString = u"Bébé"
        Main file with |$v| and eacute in the template é"""

        template = self.createAndCompile(source)
        template.v = u'Unicode String'

        assert unicode(template())


class JBQ_UTF8_Test8(CommandLineTest):
    def testStaticCompile(self):
        source = """#encoding utf-8
#set $someUnicodeString = u"Bébé"
$someUnicodeString"""

        template = self.createAndCompile(source)()

        a = unicode(template)
        if PY2:
            a = a.encode("utf-8")
        self.assertEqual("Bébé", a)

    def testDynamicCompile(self):
        source = """#encoding utf-8
#set $someUnicodeString = u"Bébé"
$someUnicodeString"""

        template = Template(source=source)

        a = unicode(template)
        if PY2:
            a = a.encode("utf-8")
        self.assertEqual("Bébé", a)


class EncodeUnicodeCompatTest(unittest.TestCase):
    """
        Taken initially from Red Hat's bugzilla #529332
        https://bugzilla.redhat.com/show_bug.cgi?id=529332
    """
    def runTest(self):
        t = Template("""Foo ${var}""", filter='EncodeUnicode')
        t.var = u"Text with some non-ascii characters: åäö"

        rc = t.respond()
        assert isinstance(rc, unicode), \
            ('Template.respond() should return unicode', rc)

        rc = str(t)
        assert isinstance(rc, str), \
            ('Template.__str__() should return a UTF-8 encoded string', rc)


class Unicode_in_SearchList_Test(CommandLineTest):
    def test_BasicASCII(self):
        source = '''This is $adjective'''

        template = self.createAndCompile(source)
        assert template and issubclass(template, Template)
        template = template(searchList=[{'adjective': u'neat'}])
        assert template.respond()

    def test_Thai(self):
        # The string is something in Thai
        source = '''This is $foo $adjective'''
        template = self.createAndCompile(source)
        assert template and issubclass(template, Template)
        template = template(
            searchList=[{
                'foo': 'bar',
                'adjective':
                    u'\u0e22\u0e34\u0e19\u0e14\u0e35\u0e15'
                    u'\u0e49\u0e2d\u0e19\u0e23\u0e31\u0e1a'
            }])
        assert template.respond()

    def test_Thai_utf8(self):
        utf8 = '\xe0\xb8\xa2\xe0\xb8\xb4\xe0\xb8\x99\xe0' \
            '\xb8\x94\xe0\xb8\xb5\xe0\xb8\x95\xe0\xb9\x89\xe0' \
            '\xb8\xad\xe0\xb8\x99\xe0\xb8\xa3\xe0\xb8\xb1\xe0\xb8\x9a'

        source = '''This is $adjective'''
        template = self.createAndCompile(source)
        assert template and issubclass(template, Template)
        template = template(searchList=[{'adjective': utf8}])
        assert template.respond()


class InlineSpanishTest(unittest.TestCase):
    def setUp(self):
        super(InlineSpanishTest, self).setUp()
        self.template = '''
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
    <title>Pagina del vendedor</title>
  </head>
  <body>
    $header
    <h2>Bienvenido $nombre.</h2>
    <br /><br /><br />
    <center>
      Usted tiene $numpedidos_noconf <a href="">pedidós</a> sin confirmar.
      <br /><br />
      Bodega tiene fecha para $numpedidos_bodega <a href="">pedidos</a>.
    </center>
  </body>
</html>
        '''  # noqa

    if PY2:  # In PY3 templates are already unicode
        def test_failure(self):
            """ Test a template lacking a proper #encoding tag """
            self.assertRaises(UnicodeDecodeError, Template, self.template,
                              searchList=[{'header': '',
                                           'nombre': '',
                                           'numpedidos_bodega': '',
                                           'numpedidos_noconf': ''}])

    def test_success(self):
        """ Test a template with a proper #encoding tag """
        template = '#encoding utf-8\n%s' % self.template
        template = Template(template, searchList=[{'header': '',
                                                   'nombre': '',
                                                   'numpedidos_bodega': '',
                                                   'numpedidos_noconf': ''}])
        self.assertTrue(unicode(template))


class CompilerTest(unittest.TestCase):
    def test_compiler_str(self):
        """ Test Compiler.__str__ """
        source = """#encoding utf-8
#set $someUnicodeString = u"Bébé"
$someUnicodeString"""
        compiler = Compiler(source)
        self.assertIsInstance(str(compiler), str)
        self.assertEqual(compiler.getModuleEncoding(), 'utf-8')



import os.path
import string
from .compat import unicode

letters = None
try:
    letters = string.ascii_letters
except AttributeError:
    letters = string.letters

_l = ['_'] * 256
for c in string.digits + letters:
    _l[ord(c)] = c
_pathNameTransChars = ''.join(_l)
del _l, c


def convertTmplPathToModuleName(tmplPath,
                                _pathNameTransChars=_pathNameTransChars,
                                splitdrive=os.path.splitdrive,
                                ):
    try:
        moduleName = splitdrive(tmplPath)[1].translate(_pathNameTransChars)
    except (UnicodeError, TypeError):
        moduleName = unicode(splitdrive(tmplPath)[1])\
            .translate(unicode(_pathNameTransChars))
    return moduleName



from glob import glob
import os
import shutil
import sys
import unittest
import Cheetah.ImportHooks
from Cheetah.compat import PY2


ImportHooksTemplatesDir = os.path.join(
    os.path.dirname(__file__), 'ImportHooksTemplates')


def setUpModule():
    sys.path.append(ImportHooksTemplatesDir)


def tearDownModule():
    assert sys.path[-1] == ImportHooksTemplatesDir
    del sys.path[-1]


def _cleanup():
    py_files = os.path.join(ImportHooksTemplatesDir, '*.py')
    pyc_files = py_files + 'c'
    for fname in glob(py_files) + glob(pyc_files):
        os.remove(fname)
    __pycache__ = os.path.join(ImportHooksTemplatesDir, '__pycache__')
    if os.path.isdir(__pycache__):
        shutil.rmtree(__pycache__)

    for modname in list(sys.modules.keys()):
        if '.ImportHooksTemplates.' in modname \
                or modname.endswith('.ImportHooksTemplates'):
            del sys.modules[modname]

    for modname in 'index', 'layout':
        if modname in sys.modules:
            del sys.modules[modname]

    Cheetah.ImportHooks.uninstall()


def _exec(code, _dict):
    exec(code, _dict)


class ImportHooksTest(unittest.TestCase):
    def setUp(self):
        _cleanup()

    def tearDown(self):
        _cleanup()

    def test_CheetahDirOwner(self):
        templates = list(sorted(os.listdir(ImportHooksTemplatesDir)))
        self.assertListEqual(templates, ['index.tmpl', 'layout.tmpl'])

        cdo = Cheetah.ImportHooks.CheetahDirOwner(ImportHooksTemplatesDir)
        index_mod = cdo.getmod('index')
        templates = os.listdir(ImportHooksTemplatesDir)
        self.assertIn('index.py', templates)
        self.assertNotIn('layout.py', templates)

        index_co = index_mod.__co__
        del index_mod.__co__
        self.assertRaises(ImportError, _exec, index_co, index_mod.__dict__)

        cdo.getmod('layout')  # Compiled to layout.py and .pyc
        self.assertIn('layout.py', os.listdir(ImportHooksTemplatesDir))

    def test_ImportHooks(self):
        templates = os.listdir(ImportHooksTemplatesDir)
        self.assertNotIn('index.py', templates)
        self.assertNotIn('layout.py', templates)
        Cheetah.ImportHooks.install()
        from index import index  # noqa
        templates = os.listdir(ImportHooksTemplatesDir)
        self.assertIn('index.py', templates)
        self.assertIn('layout.py', templates)

    def test_import_builtin(self):
        Cheetah.ImportHooks.install()
        for nm in sys.builtin_module_names:
            if nm not in sys.modules:
                __import__(nm)
                return
        raise self.fail("All builtin modules are imported")

    # _bootlocale was removed in Python 3.10:
    # https://bugs.python.org/issue42208
    if not PY2 and sys.version_info < (3, 10):
        def test_import_bootlocale(self):
            if '_bootlocale' in sys.modules:
                del sys.modules['_bootlocale']
            Cheetah.ImportHooks.install()
            import _bootlocale  # noqa: F401 '_bootlocale' imported but unused



import unittest
import Cheetah.Template
import Cheetah.Filters


class BasicMarkdownFilterTest(unittest.TestCase):
    '''
        Test that our markdown filter works
    '''
    def test_BasicHeader(self):
        template = '''
#from Cheetah.Filters import Markdown
#transform Markdown
$foo

Header
======
        '''
        expected = '''<p>bar</p>
<h1>Header</h1>'''
        try:
            template = Cheetah.Template.Template(
                template, searchList=[{'foo': 'bar'}])
            template = str(template)
            assert template == expected
        except ImportError as ex:
            print('>>> We probably failed to import markdown, bummer %s' % ex)
            return
        except Exception:
            raise


class BasicCodeHighlighterFilterTest(unittest.TestCase):
    '''
        Test that our code highlighter filter works
    '''
    def test_Python(self):
        template = '''
#from Cheetah.Filters import CodeHighlighter
#transform CodeHighlighter

def foo(self):
    return '$foo'
        '''
        template = Cheetah.Template.Template(
            template, searchList=[{'foo': 'bar'}])
        template = str(template)
        assert template, (template, 'We should have some content here...')

    def test_Html(self):
        template = '''
#from Cheetah.Filters import CodeHighlighter
#transform CodeHighlighter

<html><head></head><body>$foo</body></html>
        '''
        template = Cheetah.Template.Template(
            template, searchList=[{'foo': 'bar'}])
        template = str(template)
        assert template, (template, 'We should have some content here...')



# -------Main Package Settings-----------#
from setuptools import Extension
import os
import os.path
import sys

from Cheetah.Version import Version as version
from Cheetah.compat import PY2

# Cheetah3 has already been taken at PyPI,
# CheetahTemplate3 seems to be too long.
# CT3 is just right!
name = 'CT3'
description = "Cheetah is a template engine and code generation tool"
license = "MIT"
author = "Tavis Rudd"
author_email = "tavis@damnsimple.com"
maintainer = "Oleg Broytman"
maintainer_email = "phd@phdru.name"
url = "https://cheetahtemplate.org/"
dev_tag = ""
download_url = "https://pypi.org/project/%s/%s%s" \
    % (name, version, dev_tag)
project_urls = {
    'Homepage': 'https://cheetahtemplate.org/',
    'Documentation': 'https://cheetahtemplate.org/users_guide/index.html',
    'Download': 'https://pypi.org/project/%s/%s%s/'
    % (name, version, dev_tag),
    'Github repo': 'https://github.com/CheetahTemplate3',
    'Issue tracker': 'https://github.com/CheetahTemplate3/cheetah3/issues',
    'Wikipedia': 'https://en.wikipedia.org/wiki/CheetahTemplate',
}
del dev_tag
keywords = ["template"]
platforms = "Any"
packages = ['Cheetah',
            'Cheetah.Macros',
            'Cheetah.Templates',
            'Cheetah.Tests',
            'Cheetah.Tools',
            'Cheetah.Utils',
            ]
classifiers = [line.strip() for line in '''\
  Development Status :: 3 - Alpha
  Intended Audience :: Developers
  Intended Audience :: System Administrators
  License :: OSI Approved :: MIT License
  Operating System :: OS Independent
  Programming Language :: Python :: Implementation :: CPython
  Programming Language :: Python :: Implementation :: PyPy
  Programming Language :: Python
  Programming Language :: Python :: 2
  Programming Language :: Python :: 2.7
  Programming Language :: Python :: 3
  Programming Language :: Python :: 3.4
  Programming Language :: Python :: 3.5
  Programming Language :: Python :: 3.6
  Programming Language :: Python :: 3.7
  Programming Language :: Python :: 3.8
  Programming Language :: Python :: 3.9
  Programming Language :: Python :: 3.10
  Programming Language :: Python :: 3.11
  Topic :: Internet :: WWW/HTTP
  Topic :: Internet :: WWW/HTTP :: Dynamic Content
  Topic :: Internet :: WWW/HTTP :: Site Management
  Topic :: Software Development :: Code Generators
  Topic :: Software Development :: Libraries :: Python Modules
  Topic :: Software Development :: User Interfaces
  Topic :: Text Processing'''.splitlines() if not line.strip().startswith('#')]

if PY2:
    del line
del PY2  # Hide it from setup()

ext_modules = [
    Extension("Cheetah._namemapper",
              [os.path.join('Cheetah', 'c', '_namemapper.c')]),
    # Extension("Cheetah._verifytype",
    #            [os.path.join('Cheetah', 'c', '_verifytype.c')]),
    # Extension("Cheetah._filters",
    #            [os.path.join('Cheetah', 'c', '_filters.c')]),
    # Extension('Cheetah._template',
    #            [os.path.join('Cheetah', 'c', '_template.c')]),
]

# Data Files and Scripts
scripts = ('bin/cheetah-compile',
           'bin/cheetah',
           'bin/cheetah-analyze',
           )

data_files = ['recursive: Cheetah *.tmpl *.txt *.rst LICENSE README.rst TODO']

try:
    if sys.platform == 'win32':
        # use 'entry_points' instead of 'scripts'
        del scripts
        entry_points = {
            'console_scripts': [
                'cheetah = Cheetah.CheetahWrapper:_cheetah',
                'cheetah-compile = '
                'Cheetah.CheetahWrapper:_cheetah_compile',
            ]
        }
except ImportError:
    pass

long_description = '''\
Cheetah3 is a free and open source template engine and code generation tool.

It can be used standalone or combined with other tools and frameworks. Web
development is its principle use, but Cheetah is very flexible and
is also being used to generate C++ game code, Java, sql, form emails
and even Python code.

It's a fork of the original CheetahTemplate library.

Documentation
================================================================================
For a high-level introduction to Cheetah please refer to the User\'s Guide
at https://cheetahtemplate.org/users_guide/index.html

Credits
================================================================================
https://cheetahtemplate.org/authors.html

https://github.com/CheetahTemplate3/cheetah3/blob/master/LICENSE

Recent Changes
================================================================================
See https://cheetahtemplate.org/news.html for full details

'''

long_description_content_type = "text/x-rst"

python_requires = '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*'

extras_require = {
    'filters': ['markdown'],
    'markdown': ['markdown'],
}



#!/usr/bin/env python
"""
Core module of Cheetah's Unit-testing framework

TODO
================================================================================
# combo tests
# negative test cases for expected exceptions
# black-box vs clear-box testing
# do some tests that run the Template for long enough to check
# that the refresh code works
"""

import sys

args_l = len(sys.argv)
if args_l == 1:
    pass
elif args_l == 2 and sys.argv[1] == '--namemapper-pure':
    try:
        from Cheetah import _namemapper  # noqa
    except ImportError:
        # _namemapper hasn't been compiled so Tests/NameMapper.py
        # tests pure-python NameMapper.py; no need to duplicate these tests.
        print('Ok')
        sys.exit(0)
    sys.modules['Cheetah._namemapper'] = None
    sys._cheetah_namemapper_pure = True
else:
    sys.exit('Wrong argument or wrong number of arguments')

import unittest  # noqa: E402 module level import not at top of file

from Cheetah.Tests import Analyzer  # noqa: E402
from Cheetah.Tests import CheetahWrapper  # noqa: E402
from Cheetah.Tests import Filters  # noqa: E402
from Cheetah.Tests import ImportHooks  # noqa: E402
from Cheetah.Tests import LoadTemplate  # noqa: E402
from Cheetah.Tests import Misc  # noqa: E402
from Cheetah.Tests import NameMapper  # noqa: E402
from Cheetah.Tests import Parser  # noqa: E402
from Cheetah.Tests import Regressions  # noqa: E402
from Cheetah.Tests import SyntaxAndOutput  # noqa: E402
from Cheetah.Tests import Template  # noqa: E402
from Cheetah.Tests import TemplateCmdLineIface  # noqa: E402
from Cheetah.Tests import Unicode  # noqa: E402

SyntaxAndOutput.install_eols()

suites = [
    unittest.defaultTestLoader.loadTestsFromModule(Analyzer),
    unittest.defaultTestLoader.loadTestsFromModule(Filters),
    unittest.defaultTestLoader.loadTestsFromModule(ImportHooks),
    unittest.defaultTestLoader.loadTestsFromModule(LoadTemplate),
    unittest.defaultTestLoader.loadTestsFromModule(Misc),
    unittest.defaultTestLoader.loadTestsFromModule(NameMapper),
    unittest.defaultTestLoader.loadTestsFromModule(Parser),
    unittest.defaultTestLoader.loadTestsFromModule(Regressions),
    unittest.defaultTestLoader.loadTestsFromModule(SyntaxAndOutput),
    unittest.defaultTestLoader.loadTestsFromModule(Template),
    unittest.defaultTestLoader.loadTestsFromModule(TemplateCmdLineIface),
    unittest.defaultTestLoader.loadTestsFromModule(Unicode),
]

if not sys.platform.startswith('java'):
    suites.append(
        unittest.defaultTestLoader.loadTestsFromModule(CheetahWrapper))

if __name__ == '__main__':
    if 'xml' in sys.argv:
        from Cheetah.Tests import xmlrunner
        runner = xmlrunner.XMLTestRunner(filename='Cheetah-Tests.xml')
    else:
        runner = unittest.TextTestRunner()

    results = runner.run(unittest.TestSuite(suites))
    if results.wasSuccessful():
        sys.exit(0)
    else:
        sys.exit(1)



#!/usr/bin/env python
"""This module supports Cheetah's optional NameMapper syntax.

Overview
================================================================================

NameMapper provides a simple syntax for accessing Python data structures,
functions, and methods from Cheetah. It's called NameMapper because it 'maps'
simple 'names' in Cheetah templates to possibly more complex syntax in Python.

Its purpose is to make working with Cheetah easy for non-programmers.
Specifically, non-programmers using Cheetah should NOT need to be taught (a)
what the difference is between an object and a dictionary, (b) what functions
and methods are, and (c) what 'self' is.  A further aim (d) is to buffer the
code in Cheetah templates from changes in the implementation of the Python data
structures behind them.

Consider this scenario:

You are building a customer information system. The designers with you want to
use information from your system on the client's website --AND-- they want to
understand the display code and so they can maintian it themselves.

You write a UI class with a 'customers' method that returns a dictionary of all
the customer objects.  Each customer object has an 'address' method
that returns the a dictionary with information about the customer's address.
The designers want to be able to access that information.

Using PSP, the display code for the website would look something like the
following, assuming your servlet subclasses the class you created for managing
customer information:

  <%= self.customer()[ID].address()['city'] %>   (42 chars)

Using Cheetah's NameMapper syntax it could be any of the following:

   $self.customers()[$ID].address()['city']       (39 chars)
   --OR--
   $customers()[$ID].address()['city']
   --OR--
   $customers()[$ID].address().city
   --OR--
   $customers()[$ID].address.city
   --OR--
   $customers()[$ID].address.city
   --OR--
   $customers[$ID].address.city                   (27 chars)


Which of these would you prefer to explain to the designers, who have no
programming experience?  The last form is 15 characters shorter than the PSP
and, conceptually, is far more accessible. With PHP or ASP, the code would be
even messier than the PSP

This is a rather extreme example and, of course,
you could also just implement '$getCustomer($ID).city'
and obey the Law of Demeter (search Google for more on that).
But good object orientated design isn't the point here.

Details
================================================================================
The parenthesized letters below correspond to the aims in the second paragraph.

DICTIONARY ACCESS (a)
---------------------

NameMapper allows access to items in a dictionary
using the same dotted notation used to access object attributes in Python.
This aspect of NameMapper is known as 'Unified Dotted Notation'.

For example, with Cheetah it is possible to write::

    $customers()['kerr'].address()  --OR--  $customers().kerr.address()

where the second form is in NameMapper syntax.

This only works with dictionary keys that are also valid python identifiers::

    regex = '[a-zA-Z_][a-zA-Z_0-9]*'


AUTOCALLING (b,d)
-----------------

NameMapper automatically detects functions and methods in Cheetah $vars
and calls them if the parentheses have been left off.

For example if 'a' is an object, 'b' is a method::

  $a.b

is equivalent to::

  $a.b()

If b returns a dictionary, then following variations are possible::

  $a.b.c  --OR--  $a.b().c  --OR--  $a.b()['c']

where 'c' is a key in the dictionary that a.b() returns.

Further notes:

* NameMapper autocalls the function or method without any arguments.  Thus
  autocalling can only be used with functions or methods that either have no
  arguments or have default values for all arguments.

* NameMapper only autocalls functions and methods.
  Classes and callable object instances will not be autocalled.

* Autocalling can be disabled using Cheetah's 'useAutocalling' setting.

LEAVING OUT 'self' (c,d)
------------------------

NameMapper makes it possible to access the attributes of a servlet in Cheetah
without needing to include 'self' in the variable names.  See the NAMESPACE
CASCADING section below for details.

NAMESPACE CASCADING (d)
-----------------------
...

Implementation details
================================================================================

* NameMapper's search order is dictionary keys then object attributes

* NameMapper.NotFound is raised if a value can't be found for a name.

Performance and the C version
================================================================================

Cheetah comes with both a C version and a Python version of NameMapper.  The C
version is significantly faster and the exception tracebacks are much easier to
read.  It's still slower than standard Python syntax, but you won't notice the
difference in realistic usage scenarios.

Cheetah uses the optimized C version (_namemapper.c) if it has
been compiled or falls back to the Python version if not.
"""

import inspect
from pprint import pformat
import sys

from .compat import PY2
if PY2:
    from collections import Mapping
else:
    from collections.abc import Mapping

_INCLUDE_NAMESPACE_REPR_IN_NOTFOUND_EXCEPTIONS = False
_ALLOW_WRAPPING_OF_NOTFOUND_EXCEPTIONS = True
__all__ = ['NotFound',
           'hasKey',
           'valueForKey',
           'valueForName',
           'valueFromSearchList',
           'valueFromFrameOrSearchList',
           'valueFromFrame',
           ]

if PY2 and not hasattr(inspect.imp, 'get_suffixes'):
    # This is to fix broken behavior of the inspect module under the
    # Google App Engine
    setattr(inspect.imp, 'get_suffixes', lambda: [('.py', 'U', 1)])

# N.B. An attempt is made at the end of this module to import C versions of
# these functions.  If _namemapper.c has been compiled succesfully and the
# import goes smoothly, the Python versions defined here will be replaced with
# the C versions.


def _raiseNotFoundException(key, namespace):
    excString = "cannot find '%s'" % key
    if _INCLUDE_NAMESPACE_REPR_IN_NOTFOUND_EXCEPTIONS:
        excString += ' in the namespace %s' % pformat(namespace)
    raise NotFound(excString)


def _wrapNotFoundException(exc, fullName, namespace):
    if not _ALLOW_WRAPPING_OF_NOTFOUND_EXCEPTIONS:
        raise
    else:
        excStr = exc.args[0]
        if excStr.find('while searching') == -1:  # only wrap once!
            excStr += " while searching for '%s'" % fullName
            if _INCLUDE_NAMESPACE_REPR_IN_NOTFOUND_EXCEPTIONS:
                excStr += ' in the namespace %s' % pformat(namespace)
            exc.args = (excStr,)
        raise


def _isInstanceOrClass(obj):
    if isinstance(obj, type):
        # oldstyle
        return True

    if hasattr(obj, "__class__"):
        # newstyle
        if hasattr(obj, 'mro'):
            # type/class
            return True
        elif (hasattr(obj, '__func__') or hasattr(obj, '__code__')
                or hasattr(obj, '__self__')):
            # method, func, or builtin func
            return False
        elif hasattr(obj, '__init__'):
            # instance
            return True
    return False


def hasKey(obj, key):
    """Determine if 'obj' has 'key' """
    if isinstance(obj, Mapping) and key in obj:
        return True
    elif hasattr(obj, key):
        return True
    else:
        return False


def _valueForName(obj, name, executeCallables=False):
    nameChunks = name.split('.')
    for key in nameChunks:
        if isinstance(obj, Mapping) and key in obj:
            nextObj = obj[key]
        else:
            try:
                nextObj = getattr(obj, key)
            except AttributeError:
                _raiseNotFoundException(key, obj)

        if executeCallables and callable(nextObj) and \
                not _isInstanceOrClass(nextObj):
            obj = nextObj()
        else:
            obj = nextObj
    return obj


def _namespaces(callerFrame, searchList=None):
    yield callerFrame.f_locals
    if searchList:
        for namespace in searchList:
            yield namespace
    yield callerFrame.f_globals
    yield __builtins__


def hasName(obj, name):
    # Not in the C version
    """Determine if 'obj' has the 'name' """
    key = name.split('.')[0]
    if not hasKey(obj, key):
        return False
    try:
        valueForName(obj, name)
        return True
    except NotFound:
        return False


try:
    if getattr(sys, '_cheetah_namemapper_pure', False):
        raise ValueError  # Not an error, just a way to skip ``_namemapper``
    from ._namemapper import NotFound, valueForKey, valueForName, \
        valueFromSearchList, valueFromFrameOrSearchList, valueFromFrame
    C_VERSION = True
except Exception:
    class NotFound(LookupError):
        pass

    def valueForKey(obj, key):
        if isinstance(obj, Mapping) and key in obj:
            return obj[key]
        elif hasattr(obj, key):
            return getattr(obj, key)
        else:
            _raiseNotFoundException(key, obj)

    def valueForName(obj, name, executeCallables=False):
        try:
            return _valueForName(obj, name, executeCallables)
        except NotFound as e:
            _wrapNotFoundException(e, fullName=name, namespace=obj)

    def valueFromSearchList(searchList, name, executeCallables=False):
        key = name.split('.')[0]
        for namespace in searchList:
            if hasKey(namespace, key):
                return _valueForName(namespace, name,
                                     executeCallables=executeCallables)
        _raiseNotFoundException(key, searchList)

    def valueFromFrame(name, executeCallables=False, frame=None):
        # @@TR consider implementing the C version the same way
        # at the moment it provides a seperate but mirror implementation
        # to valueFromFrameOrSearchList
        try:
            if not frame:
                frame = inspect.stack()[1][0]
            return valueFromFrameOrSearchList(
                searchList=None, name=name,
                executeCallables=executeCallables, frame=frame)
        finally:
            del frame

    def valueFromFrameOrSearchList(searchList, name, executeCallables=False,
                                   frame=None):
        def __valueForName():
            try:
                return _valueForName(namespace, name,
                                     executeCallables=executeCallables)
            except NotFound as e:
                _wrapNotFoundException(e, fullName=name, namespace=searchList)
        try:
            if not frame:
                frame = inspect.stack()[1][0]
            key = name.split('.')[0]
            for namespace in _namespaces(frame, searchList):
                if hasKey(namespace, key):
                    return __valueForName()
            _raiseNotFoundException(key, searchList)
        finally:
            del frame

    # It is possible with PyPy, Jython or Windows, for example,
    # that _namemapper.c hasn't been compiled.
    C_VERSION = False


##################################################
# CLASSES


class Mixin:
    """@@ document me"""
    def valueForName(self, name):
        return valueForName(self, name)

    def valueForKey(self, key):
        return valueForKey(self, key)

##################################################
# if run from the command line ##


def example():
    class A(Mixin):
        classVar = 'classVar val'

        def method(self, arg='method 1 default arg'):
            return arg

        def method2(self, arg='meth 2 default arg'):
            return {'item1': arg}

        def method3(self, arg='meth 3 default'):
            return arg

    class B(A):
        classBvar = 'classBvar val'

    a = A()
    a.one = 'valueForOne'

    def function(whichOne='default'):
        values = {
            'default': 'default output',
            'one': 'output option one',
            'two': 'output option two'
        }
        return values[whichOne]

    a.dic = {
        'func': function,
        'method': a.method3,
        'item': 'itemval',
        'subDict': {'nestedMethod': a.method3}
    }

    b = 'this is local b'  # noqa: F841 local variable assigned but never used

    print(valueForKey(a.dic, 'subDict'))
    print(valueForName(a, 'dic.item'))
    print(valueForName(vars(), 'b'))
    print(valueForName(__builtins__, 'dir')())
    print(valueForName(vars(), 'a.classVar'))
    print(valueForName(vars(), 'a.dic.func', executeCallables=True))
    print(valueForName(vars(), 'a.method2.item1', executeCallables=True))


if __name__ == '__main__':
    example()



#!/usr/bin/env python

import os
import pprint

from . import Parser
from . import Compiler
from . import Template


class Analyzer(Parser.Parser):
    def __init__(self, *args, **kwargs):
        self.calls = {}
        super(Analyzer, self).__init__(*args, **kwargs)

    def eatDirective(self):
        directive = self.matchDirective()
        try:
            self.calls[directive] += 1
        except KeyError:
            self.calls[directive] = 1
        super(Analyzer, self).eatDirective()


class AnalysisCompiler(Compiler.ModuleCompiler):
    parserClass = Analyzer


def analyze(source):
    klass = Template.Template.compile(source, compilerClass=AnalysisCompiler)
    return klass._CHEETAH_compilerInstance._parser.calls


def main_file(f):
    fd = open(f, 'r')
    try:
        print(u'>>> Analyzing %s' % f)
        calls = analyze(fd.read())
        return calls
    finally:
        fd.close()


def _find_templates(directory, suffix):
    for root, dirs, files in os.walk(directory):
        for f in files:
            if not f.endswith(suffix):
                continue
            yield root + os.path.sep + f


def _analyze_templates(iterable):
    for template in iterable:
        yield main_file(template)


def main_dir(opts):
    results = _analyze_templates(_find_templates(opts.dir, opts.suffix))
    totals = {}
    for series in results:
        if not series:
            continue
        for k, v in series.items():
            try:
                totals[k] += v
            except KeyError:
                totals[k] = v
    return totals


def main():
    from optparse import OptionParser
    op = OptionParser()
    op.add_option('-f', '--file', dest='file', default=None,
                  help='Specify a single file to analyze')
    op.add_option('-d', '--dir', dest='dir', default=None,
                  help='Specify a directory of templates to analyze')
    op.add_option('--suffix', default='tmpl', dest='suffix',
                  help='Specify a custom template file suffix '
                  'for the -d option (default: "tmpl")')
    opts, args = op.parse_args()

    if not opts.file and not opts.dir:
        op.print_help()
        return

    results = None
    if opts.file:
        results = main_file(opts.file)
    if opts.dir:
        results = main_dir(opts)

    pprint.pprint(results)


if __name__ == '__main__':
    main()



import os
from ..cheetahsupport import TurboCheetah

here = os.path.dirname(__file__)

values = {
    'v': 'VV',
    'one': 1,
}


def test_normal():
    plugin = TurboCheetah()
    # Make sure a simple test works:
    s = plugin.render(values, template='turbocheetah.tests.simple1')
    assert s.strip() == 'This is a test: VV'
    # Make sure one template can inherit from another:
    s = plugin.render(values, template='turbocheetah.tests.import_inherit')
    assert s.strip() == 'Inherited: import'


def test_path():
    plugin = TurboCheetah()
    plugin.search_path = [here]
    # Make sure we pick up filenames (basic test):
    s = plugin.render(values, template_file='simple1')
    assert s.strip() == 'This is a test: VV'
    # Make sure we pick up subdirectories:
    s = plugin.render(values, template_file='sub/master')
    assert s.strip() == 'sub1: 1'


def test_search():
    plugin = TurboCheetah()
    plugin.search_path = [os.path.join(here, 'sub'),
                          os.path.join(here, 'sub2'),
                          here]
    # Pick up from third entry:
    s = plugin.render(values, template_file='simple1')
    assert s.strip() == 'This is a test: VV'
    # Pick up from sub/master, non-ambiguous:
    s = plugin.render(values, template_file='master')
    assert s.strip() == 'sub1: 1'
    # Pick up from sub/page, inherit from sub/template:
    s = plugin.render(values, template_file='page')
    assert s.strip() == 'SUB: sub content'
    # Pick up from sub2/page_over, inherit from sub/template:
    s = plugin.render(values, template_file='page_over')
    assert s.strip() == 'SUB: override content'
    # Pick up from sub/page_template_over, inherit from
    # sub2/template_over:
    s = plugin.render(values, template_file='page_template_over')
    assert s.strip() == 'OVER: sub content'
    # Change page, make sure that undoes overrides:
    plugin.search_path = [os.path.join(here, 'sub'),
                          here]
    s = plugin.render(values, template_file='page_over')
    assert s.strip() == 'SUB: sub content'


def test_string():
    # Make sure simple string evaluation works:
    plugin = TurboCheetah()
    s = plugin.render(values, template_string="""Hey $v""")
    assert s == "Hey VV"
    # Make sure a string can inherit from a file:
    plugin.search_path = [here]
    s = plugin.render(values, template_string="#extends inherit_from\ns value")
    assert s.strip() == 'inherit: s value'



#!/usr/bin/env python
import os

try:
    os.remove('MANIFEST')  # to avoid those bloody out-of-date manifests!!
except Exception:
    pass

import SetupTools
import SetupConfig
configurations = (SetupConfig,)
SetupTools.run_setup(configurations)



#!/usr/bin/env python

Version = '3.3.1'
VersionTuple = (3, 3, 1, 'final', 0)

MinCompatibleVersion = '3.0.0a1'
MinCompatibleVersionTuple = (3, 0, 0, 'alpha', 1)

####


def convertVersionStringToTuple(s):
    versionNum = [0, 0, 0]
    releaseType = 'final'
    releaseTypeSubNum = 0
    if s.find('a') != -1:
        num, releaseTypeSubNum = s.split('a')
        releaseType = 'alpha'
    elif s.find('b') != -1:
        num, releaseTypeSubNum = s.split('b')
        releaseType = 'beta'
    elif s.find('rc') != -1:
        num, releaseTypeSubNum = s.split('rc')
        releaseType = 'candidate'
    else:
        num = s
    num = num.split('.')
    for i in range(len(num)):
        versionNum[i] = int(num[i])
    if len(versionNum) < 3:
        versionNum += [0]
    releaseTypeSubNum = int(releaseTypeSubNum)

    return tuple(versionNum + [releaseType, releaseTypeSubNum])


if __name__ == '__main__':
    c = convertVersionStringToTuple
    print(c('2.0a1'))
    print(c('2.0b1'))
    print(c('2.0rc1'))
    print(c('2.0'))
    print(c('2.0.2'))

    assert c('0.9.19b1') < c('0.9.19')
    assert c('0.9b1') < c('0.9.19')

    assert c('2.0a2') > c('2.0a1')
    assert c('2.0b1') > c('2.0a2')
    assert c('2.0b2') > c('2.0b1')
    assert c('2.0b2') == c('2.0b2')

    assert c('2.0rc1') > c('2.0b1')
    assert c('2.0rc2') > c('2.0rc1')
    assert c('2.0rc2') > c('2.0b1')

    assert c('2.0') > c('2.0a1')
    assert c('2.0') > c('2.0b1')
    assert c('2.0') > c('2.0rc1')
    assert c('2.0.1') > c('2.0')
    assert c('2.0rc1') > c('2.0b1')
